from openwisp_users.tests.test_admin import TestBasicUsersIntegration


class TestUsersIntegration(TestBasicUsersIntegration):
    """
    tests integration with openwisp_users
    """

    is_integration_test = True

    _notifications_params = {
        'notificationsetting_set-TOTAL_FORMS': 0,
        'notificationsetting_set-INITIAL_FORMS': 0,
        'notificationsetting_set-MIN_NUM_FORMS': 0,
        'notificationsetting_set-MAX_NUM_FORMS': 0,
    }

    def _get_org_edit_form_inline_params(self, user, org):
        params = super()._get_org_edit_form_inline_params(user, org)
        params.update(self._notifications_params)
        return params

    def _get_user_edit_form_inline_params(self, user, org):
        params = super()._get_user_edit_form_inline_params(user, org)
        params.update(self._notifications_params)
        return params


del TestBasicUsersIntegration



from asgiref.sync import async_to_sync
from channels import layers
from django.dispatch import Signal

update_topology = Signal()
update_topology.__doc__ = """
Providing arguments: ['topology']
"""


def broadcast_topology(topology, *args, **kwargs):
    channel_layer = layers.get_channel_layer()
    async_to_sync(channel_layer.group_send)(
        f'topology-{topology.pk}',
        {'type': 'send_topology_update', 'data': topology.json()},
    )



from unittest.mock import patch
from uuid import uuid4

import swapper
from django.contrib.auth.models import Permission
from django.test import TestCase
from django.urls import reverse
from rest_framework.views import APIView

from openwisp_network_topology.tasks import handle_update_topology
from openwisp_users.tests.utils import TestOrganizationMixin
from openwisp_utils.tests import AssertNumQueriesSubTestMixin

from .utils import CreateGraphObjectsMixin, LoadMixin, UnpublishMixin

Link = swapper.load_model('topology', 'Link')
Node = swapper.load_model('topology', 'Node')
Snapshot = swapper.load_model('topology', 'Snapshot')
Topology = swapper.load_model('topology', 'Topology')
Organization = swapper.load_model('openwisp_users', 'Organization')
OrganizationUser = swapper.load_model('openwisp_users', 'OrganizationUser')
Group = swapper.load_model('openwisp_users', 'Group')


class TestApi(
    AssertNumQueriesSubTestMixin,
    CreateGraphObjectsMixin,
    UnpublishMixin,
    LoadMixin,
    TestOrganizationMixin,
    TestCase,
):
    list_url = reverse('network_collection')
    topology_model = Topology
    node_model = Node
    link_model = Link
    snapshot_model = Snapshot

    def setUp(self):
        org = self._get_org()
        self.topology = self._create_topology(organization=org)
        user = self._create_user(username='tester', email='tester@email.com')
        user.groups.set(Group.objects.filter(name='Operator'))
        perm = Permission.objects.filter(codename__endswith='topology')
        user.user_permissions.add(*perm)
        self._create_org_user(user=user, organization=org, is_admin=True)
        self.node1 = self._create_node(
            label="node1",
            addresses=["192.168.0.1"],
            topology=self.topology,
            organization=org,
        )
        self.node2 = self._create_node(
            label="node2",
            addresses=["192.168.0.2"],
            topology=self.topology,
            organization=org,
        )
        self.link = self._create_link(
            source=self.node1, target=self.node2, topology=self.topology
        )
        self.client.force_login(user)

    @property
    def detail_url(self):
        t = self.topology_model.objects.first()
        return reverse('network_graph', args=[t.pk])

    @property
    def receive_url(self):
        t = self.topology_model.objects.first()
        path = reverse('receive_topology', args=[t.pk])
        return f'{path}?key=test'

    @property
    def snapshot_url(self):
        t = self.topology_model.objects.first()
        path = reverse('network_graph_history', args=[t.pk])
        return f'{path}?date={self.snapshot_date}'

    def _set_receive(self):
        t = self.topology_model.objects.first()
        t.parser = 'netdiff.NetJsonParser'
        t.strategy = 'receive'
        t.key = 'test'
        t.expiration_time = 0
        t.save()

    @property
    def snapshot_date(self):
        self.topology_model.objects.first().save_snapshot()
        return self.snapshot_model.objects.first().date

    def test_list(self):
        response = self.client.get(self.list_url)
        self.assertEqual(response.data['type'], 'NetworkCollection')
        self.assertEqual(len(response.data['collection']), 1)

    def test_list_topology_filter(self):
        admin = self._create_admin()
        self.client.force_login(admin)
        org2 = self._create_org(name='org2')
        self._create_topology(
            organization=org2, parser='netdiff.OpenvpnParser', strategy='receive'
        )

        with self.subTest('Test without filter'):
            response = self.client.get(self.list_url)
            self.assertEqual(len(response.data['collection']), 2)

        with self.subTest('Test filter with parser'):
            response = self.client.get(f'{self.list_url}?parser=netdiff.OpenvpnParser')
            self.assertEqual(len(response.data['collection']), 1)

        with self.subTest('Test filter with organization'):
            response = self.client.get(f'{self.list_url}?organization={org2.pk}')
            self.assertEqual(len(response.data['collection']), 1)

        with self.subTest('Test filter with organization slug'):
            response = self.client.get(f'{self.list_url}?organization_slug={org2.slug}')
            self.assertEqual(len(response.data['collection']), 1)

        with self.subTest('Test filter with strategy'):
            response = self.client.get(f'{self.list_url}?strategy=receive')
            self.assertEqual(len(response.data['collection']), 1)

    def test_detail(self):
        response = self.client.get(self.detail_url)
        self.assertEqual(response.data['type'], 'NetworkGraph')

    def test_list_include_unpublished(self):
        self._unpublish()
        path = f'{self.list_url}?include_unpublished=true'
        response = self.client.get(path)
        self.assertEqual(len(response.data['collection']), 1)

        path = f'{self.list_url}?published=true'
        response = self.client.get(path)
        self.assertEqual(len(response.data['collection']), 0)

    def test_detail_unpublished_topology(self):
        self._unpublish()
        with self.subTest('Test view returns 404 for unpublished topology'):
            response = self.client.get(self.detail_url)
            self.assertEqual(response.status_code, 404)

        with self.subTest('Test "include_unpublished" filter'):
            path = f'{self.detail_url}?include_unpublished=true'
            response = self.client.get(path)
            self.assertEqual(response.status_code, 200)

    def test_receive(self):
        self._set_receive()
        self.node_model.objects.all().delete()
        data = self._load('static/netjson-1-link.json')
        response = self.client.post(self.receive_url, data, content_type='text/plain')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['detail'], 'data received successfully')
        self.assertEqual(self.node_model.objects.count(), 2)
        self.assertEqual(self.link_model.objects.count(), 1)

    @patch('openwisp_network_topology.tasks.handle_update_topology.delay')
    def test_background_topology_update_task(self, mocked_task):
        self._set_receive()
        self.node_model.objects.all().delete()
        topology = self.topology_model.objects.first()
        data = self._load('static/netjson-1-link.json')
        response = self.client.post(self.receive_url, data, content_type='text/plain')
        mocked_task.assert_called_once_with(topology.pk, topology.diff(data))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['detail'], 'data received successfully')

    @patch('openwisp_network_topology.tasks.logger.warning')
    def test_background_topology_update_task_warning(self, mock_warn):
        invalid_topology_pk = str(uuid4())
        data = {'test_key': 'test_value'}
        handle_update_topology(invalid_topology_pk, data)
        mock_warn.assert_called_once_with(
            f'handle_update_topology("{invalid_topology_pk}") failed: Topology matching query does not exist.'
        )

    @patch('openwisp_network_topology.tasks.handle_update_topology.delay')
    def test_handle_update_topology_400_unrecognized_format(self, mocked_task):
        self._set_receive()
        self.node_model.objects.all().delete()
        data = 'WRONG'
        response = self.client.post(self.receive_url, data, content_type='text/plain')
        mocked_task.assert_not_called()
        self.assertEqual(response.status_code, 400)
        self.assertIn('not recognized', response.data['detail'])

    def test_receive_404(self):
        # topology is set to FETCH strategy
        response = self.client.post(self.receive_url, content_type='text/plain')
        self.assertEqual(response.status_code, 404)

    def test_receive_415(self):
        self._set_receive()
        data = self._load('static/netjson-1-link.json')
        response = self.client.post(
            self.receive_url, data, content_type='application/xml'
        )
        self.assertEqual(response.status_code, 415)

    def test_receive_400_missing_key(self):
        self._set_receive()
        data = self._load('static/netjson-1-link.json')
        response = self.client.post(
            self.receive_url.replace('?key=test', ''), data, content_type='text/plain'
        )
        self.assertEqual(response.status_code, 400)
        self.assertIn('missing required', response.data['detail'])

    def test_receive_400_unrecognized_format(self):
        self._set_receive()
        self.node_model.objects.all().delete()
        data = 'WRONG'
        response = self.client.post(self.receive_url, data, content_type='text/plain')
        self.assertEqual(response.status_code, 400)
        self.assertIn('not recognized', response.data['detail'])

    def test_receive_403(self):
        self._set_receive()
        data = self._load('static/netjson-1-link.json')
        response = self.client.post(
            self.receive_url.replace('?key=test', '?key=wrong'),
            data,
            content_type='text/plain',
        )
        self.assertEqual(response.status_code, 403)

    def test_receive_options(self):
        self._set_receive()
        response = self.client.options(self.receive_url)
        self.assertEqual(response.data['parses'], ['text/plain'])

    def test_snapshot(self):
        response = self.client.get(self.snapshot_url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['type'], 'NetworkGraph')

    def test_snapshot_missing_date_400(self):
        date = self.snapshot_date
        response = self.client.get(
            self.snapshot_url.replace('?date={0}'.format(date), '')
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.data['detail'], 'missing required "date" parameter')

    def test_snapshot_invalid_date_403(self):
        date = self.snapshot_date
        url = self.snapshot_url.replace('?date={0}'.format(date), '?date=wrong-date')
        response = self.client.get(url)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(response.data['detail'], 'invalid date supplied')

    def test_snapshot_no_snapshot_404(self):
        date = self.snapshot_date
        url = self.snapshot_url.replace('?date={0}'.format(date), '?date=2001-01-01')
        response = self.client.get(url)
        self.assertEqual(response.status_code, 404)
        self.assertIn('no snapshot found', response.data['detail'])

    def _test_api_with_unauthenticated_user(self, url):
        self.client.logout()
        r = self.client.get(url)
        self.assertEqual(r.status_code, 403)
        self.assertEqual(
            r.data['detail'], 'Authentication credentials were not provided.'
        )
        self.assertEqual(len(r.data), 1)

    def _test_api_with_not_a_manager_user(self, user, url, has_detail=True):
        OrganizationUser.objects.filter(user=user).delete()
        perm = Permission.objects.filter(codename__endswith='topology')
        user.user_permissions.add(*perm)
        self.client.force_login(user)
        r = self.client.get(url)
        if not has_detail:
            self.assertEqual(r.status_code, 200)
            self.assertEqual(r.data['type'], 'NetworkCollection')
            self.assertEqual(len(r.data['collection']), 0)
            self.assertNotEqual(Topology.objects.all().count(), 0)
        else:
            detail = (
                'User is not a manager of the organization to '
                'which the requested resource belongs.'
            )
            self.assertEqual(r.status_code, 403)
            self.assertEqual(r.data['detail'], detail)
            self.assertEqual(len(r.data), 1)

    def _test_api_with_not_permitted_user(self, user, url):
        t = self.topology_model.objects.first()
        self._create_org_user(user=user, organization=t.organization, is_admin=True)
        self.client.force_login(user)
        r = self.client.get(url)
        self.assertEqual(r.status_code, 403)
        self.assertEqual(
            r.data['detail'], 'You do not have permission to perform this action.'
        )
        self.assertEqual(len(r.data), 1)

    def test_modelpermission_class_with_change_perm(self):
        t = self.topology_model.objects.first()
        user = self._create_user(username='list-user', email='list@email.com')
        self._create_org_user(user=user, organization=t.organization, is_admin=True)
        change_perm = Permission.objects.filter(codename='change_topology')
        user.user_permissions.add(*change_perm)
        self.client.force_login(user)
        with self.subTest('List url'):
            url = self.list_url
            with self.assertNumQueries(7):
                response = self.client.get(url)
            self.assertEqual(response.status_code, 200)
        with self.subTest('Detail url'):
            url = self.detail_url
            with self.assertNumQueries(7):
                response = self.client.get(url)
            self.assertEqual(response.status_code, 200)

    def test_modelpermission_class_with_view_perm(self):
        t = self.topology_model.objects.first()
        user = self._create_user(username='list-user', email='list@email.com')
        self._create_org_user(user=user, organization=t.organization, is_admin=True)
        view_perm = Permission.objects.filter(codename='view_topology')
        user.user_permissions.add(*view_perm)
        self.client.force_login(user)
        with self.subTest('List url'):
            url = self.list_url
            with self.assertNumQueries(7):
                response = self.client.get(url)
            self.assertEqual(response.status_code, 200)
        with self.subTest('Detail url'):
            url = self.detail_url
            with self.assertNumQueries(7):
                response = self.client.get(url)
            self.assertEqual(response.status_code, 200)

    def test_modelpermission_class_with_no_perm(self):
        t = self.topology_model.objects.first()
        user = self._create_user(username='list-user', email='list@email.com')
        self._create_org_user(user=user, organization=t.organization, is_admin=True)
        self.client.force_login(user)
        with self.subTest('List url'):
            url = self.list_url
            with self.assertNumQueries(4):
                response = self.client.get(url)
            self.assertEqual(response.status_code, 403)
        with self.subTest('Detail url'):
            url = self.detail_url
            with self.assertNumQueries(4):
                response = self.client.get(url)
            self.assertEqual(response.status_code, 403)

    def test_list_with_auth_enabled(self):
        user = self._create_user(username='list-user', email='list@email.com')
        with self.subTest('test api with unauthenticated user'):
            self._test_api_with_unauthenticated_user(self.list_url)

        with self.subTest('test api with not a permitted user'):
            self._test_api_with_not_permitted_user(user, self.list_url)

        with self.subTest('test api with not a member user'):
            self._test_api_with_not_a_manager_user(user, self.list_url)

    def test_detail_with_auth_enabled(self):
        user = self._create_user(username='detail-user', email='detail@email.com')
        with self.subTest('test api with unauthenticated user'):
            self._test_api_with_unauthenticated_user(self.detail_url)

        with self.subTest('test api with not a permitted user'):
            self._test_api_with_not_permitted_user(user, self.detail_url)

        with self.subTest('test api with not a member user'):
            self._test_api_with_not_a_manager_user(user, self.detail_url)

    def test_snapshot_with_auth_enabled(self):
        user = self._create_user(username='snapshot-user', email='snapshot@email.com')
        with self.subTest('test api with unauthenticated user'):
            self._test_api_with_unauthenticated_user(self.snapshot_url)

        with self.subTest('test api with not a permitted user'):
            self._test_api_with_not_permitted_user(user, self.snapshot_url)

        with self.subTest('test api with not a member user'):
            self._test_api_with_not_a_manager_user(user, self.snapshot_url)

    def _successful_api_tests(self):
        with self.subTest('test receive'):
            self._set_receive()
            self.node_model.objects.all().delete()
            data = self._load('static/netjson-1-link.json')
            response = self.client.post(
                self.receive_url, data, content_type='text/plain'
            )
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.data['detail'], 'data received successfully')
            self.assertEqual(self.node_model.objects.count(), 2)
            self.assertEqual(self.link_model.objects.count(), 1)

        with self.subTest('test history'):
            response = self.client.get(self.snapshot_url)
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.data['type'], 'NetworkGraph')

    @patch.object(APIView, 'get_permissions', return_value=[])
    @patch.object(APIView, 'get_authenticators', return_value=[])
    def test_api_with_auth_disabled(self, perm_mocked, auth_mocked):
        user = self._get_user(username='tester')
        self.client.logout()
        self._successful_api_tests()
        self.client.force_login(user)

    def test_superuser_with_api_auth_enabled(self):
        user = self._create_admin(username='superapi', email='superapi@email.com')
        self.client.force_login(user)
        self._successful_api_tests()

    @patch.object(APIView, 'get_permissions', return_value=[])
    @patch.object(APIView, 'get_authenticators', return_value=[])
    def test_superuser_with_api_auth_disabled(self, perm_mocked, auth_mocked):
        user = self._create_admin(username='superapi', email='superapi@email.com')
        self.client.force_login(user)
        self._successful_api_tests()

    def test_fetch_topology_create_api(self):
        path = reverse('network_collection')
        data = {
            'label': 'test-fetch-topology',
            'organization': self._get_org().id,
            'parser': 'netdiff.OlsrParser',
            'strategy': 'fetch',
            'url': 'http://127.0.0.1:9090',
            'published': True,
        }
        with self.assertNumQueries(12):
            response = self.client.post(path, data, content_type='application/json')
        self.assertEqual(response.status_code, 201)
        self.assertEqual(response.data['label'], 'test-fetch-topology')
        self.assertEqual(response.data['parser'], 'netdiff.OlsrParser')

    def test_receive_topology_create_api(self):
        path = reverse('network_collection')
        data = {
            'label': 'test-receive-topology',
            'organization': self._get_org().pk,
            'parser': 'netdiff.OlsrParser',
            'strategy': 'receive',
            'key': 'A3DJ62jhd49',
            'expiration_time': 360,
            'published': True,
        }
        with self.assertNumQueries(12):
            response = self.client.post(path, data, content_type='application/json')
        self.assertEqual(response.status_code, 201)
        self.assertEqual(response.data['label'], 'test-receive-topology')
        self.assertEqual(response.data['parser'], 'netdiff.OlsrParser')
        self.assertEqual(response.data['key'], 'A3DJ62jhd49')

    def test_topology_detail_receive_url_api(self):
        org1 = self._get_org()
        topo = self._create_topology(organization=org1)
        path = reverse('network_graph', args=(topo.pk,))
        r_path = reverse('receive_topology', args=[topo.pk])
        receive_url = 'http://testserver{0}?key={1}'.format(r_path, topo.key)
        with self.assertNumQueries(7):
            response = self.client.get(path)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['receive_url'], receive_url)

    def test_topology_receive_no_key_create_api(self):
        path = reverse('network_collection')
        data = {
            'label': 'test-receive-topology',
            'organization': self._get_org().pk,
            'parser': 'netdiff.OlsrParser',
            'strategy': 'receive',
            'key': '',
            'expiration_time': 360,
            'published': True,
        }
        with self.assertNumQueries(7):
            response = self.client.post(path, data, content_type='application/json')
        self.assertEqual(response.status_code, 400)
        self.assertIn(
            'a key must be specified when using RECEIVE strategy', str(response.content)
        )

    def test_get_topology_detail_api(self):
        org1 = self._get_org()
        topo = self._create_topology(organization=org1)
        path = reverse('network_graph', args=(topo.pk,))
        with self.assertNumQueries(7):
            response = self.client.get(path)
        self.assertEqual(response.status_code, 200)

    def test_get_topology_detail_with_link_api(self):
        path = reverse('network_graph', args=(self.topology.pk,))
        with self.assertNumQueries(7):
            response = self.client.get(path)
        self.assertEqual(response.status_code, 200)
        self.assertNotEqual(response.data['links'], [])

    def test_put_topology_detail_api(self):
        path = reverse('network_graph', args=[self.topology.pk])
        data = {
            'label': 'ChangeTestNetwork',
            'organization': self._get_org().pk,
            'parser': 'netdiff.OlsrParser',
        }
        with self.assertNumQueries(12):
            response = self.client.put(path, data, content_type='application/json')
        self.topology.refresh_from_db()
        self.assertEqual(self.topology.label, 'ChangeTestNetwork')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['label'], 'ChangeTestNetwork')
        self.assertEqual(response.data['type'], 'NetworkGraph')

    def test_change_strategy_fetch_api_400(self):
        org1 = self._get_org()
        topo = self._create_topology(organization=org1)
        path = reverse('network_graph', args=(topo.pk,))
        data = {
            'label': 'ChangeTestNetwork',
            'organization': org1.pk,
            'parser': 'netdiff.OlsrParser',
            'strategy': 'fetch',
            'url': '',
        }
        with self.assertNumQueries(6):
            response = self.client.put(path, data, content_type='application/json')
        self.assertEqual(response.status_code, 400)
        self.assertIn(
            'An url must be specified when using FETCH strategy', str(response.content)
        )

    def test_change_strategy_receive_api_400(self):
        org1 = self._get_org()
        topo = self._create_topology(organization=org1)
        path = reverse('network_graph', args=(topo.pk,))
        data = {
            'label': 'ChangeTestNetwork',
            'organization': self._get_org().pk,
            'parser': 'netdiff.OlsrParser',
            'strategy': 'receive',
            'key': '',
        }
        with self.assertNumQueries(6):
            response = self.client.put(path, data, content_type='application/json')
        self.assertEqual(response.status_code, 400)
        self.assertIn(
            'A key must be specified when using RECEIVE strategy', str(response.content)
        )

    def test_change_strategy_fetch_api_200(self):
        org1 = self._get_org()
        topo = self._create_topology(organization=org1, strategy='receive')
        path = reverse('network_graph', args=[topo.pk])
        self.assertEqual(topo.strategy, 'receive')
        data = {
            'label': 'ChangeTestNetwork',
            'organization': org1.pk,
            'parser': 'netdiff.OlsrParser',
            'strategy': 'fetch',
            'url': 'http://127.0.0.1:9090',
        }
        with self.assertNumQueries(12):
            response = self.client.put(path, data, content_type='application/json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['strategy'], 'fetch')
        topo.refresh_from_db()
        self.assertEqual(topo.strategy, 'fetch')

    def test_change_strategy_receive_api_200(self):
        org1 = self._get_org()
        topo = self._create_topology(organization=org1, strategy='fetch')
        path = reverse('network_graph', args=[topo.pk])
        self.assertEqual(topo.strategy, 'fetch')
        data = {
            'label': 'ChangeTestNetwork',
            'organization': org1.pk,
            'parser': 'netdiff.OlsrParser',
            'strategy': 'receive',
            'key': 12345,
        }
        with self.assertNumQueries(12):
            response = self.client.put(path, data, content_type='application/json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['strategy'], 'receive')
        topo.refresh_from_db()
        self.assertEqual(topo.strategy, 'receive')

    def test_patch_topology_detail_api(self):
        path = reverse('network_graph', args=(self.topology.pk,))
        data = {
            'label': 'ChangeTestNetwork',
        }
        with self.assertNumQueries(11):
            response = self.client.patch(path, data, content_type='application/json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['label'], 'ChangeTestNetwork')

    def test_delete_topology_api(self):
        path = reverse('network_graph', args=(self.topology.pk,))
        response = self.client.delete(path)
        self.assertEqual(response.status_code, 204)

    def test_topology_filter_by_org_api(self):
        org1 = self._create_org(name='org1')
        org2 = self._create_org(name='org2')
        topo1 = self._create_topology(label='topo1', organization=org1)
        topo2 = self._create_topology(label='topo2', organization=org2)
        user1 = self._create_user(username='test-filter', email='test@filter.com')
        self._create_org_user(user=user1, organization=org1, is_admin=True)
        view_perm = Permission.objects.filter(codename='view_topology')
        user1.user_permissions.add(*view_perm)
        self.client.force_login(user1)
        with self.subTest('test network collection view'):
            path = reverse('network_collection')
            with self.assertNumQueries(7):
                response = self.client.get(path)
            self.assertEqual(response.status_code, 200)
            self.assertIn(str(topo1.id), str(response.content))
            self.assertNotIn(str(topo2.id), str(response.content))

        with self.subTest('test network graph view'):
            # Get the topology graph view of member org 200
            path = reverse('network_graph', args=(topo1.pk,))
            with self.assertNumQueries(7):
                response = self.client.get(path)
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.data['id'], str(topo1.id))

            # Get the topology graph view of different org 404
            path = reverse('network_graph', args=(topo2.pk,))
            with self.assertNumQueries(5):
                response = self.client.get(path)
            self.assertEqual(response.status_code, 404)

    def test_topology_filter_fields_by_org_api(self):
        org1 = self._get_org()
        user1 = self._create_user(username='test-filter', email='test@filter.com')
        self._create_org_user(user=user1, organization=org1, is_admin=True)
        topo_perm = Permission.objects.filter(codename__endswith='topology')
        user1.user_permissions.add(*topo_perm)
        self.client.force_login(user1)
        with self.subTest('test network collection view'):
            path = reverse('network_collection')
            with self.assertNumQueries(8):
                response = self.client.get(path, {'format': 'api'})
            self.assertEqual(response.status_code, 200)
            self.assertEqual(Organization.objects.count(), 2)
            self.assertContains(response, 'test org</option>')
            self.assertNotContains(response, 'default</option>')

        with self.subTest('test network graph view'):
            topo1 = self._create_topology(label='topo1', organization=org1)
            path = reverse('network_graph', args=(topo1.pk,))
            with self.assertNumQueries(14):
                response = self.client.get(path, {'format': 'api'})
            self.assertEqual(response.status_code, 200)
            self.assertEqual(Organization.objects.count(), 2)
            self.assertContains(response, 'test org</option>')
            self.assertNotContains(response, 'default</option>')

    def test_node_list_api(self):
        path = reverse('node_list')
        with self.assertNumQueries(6):
            response = self.client.get(path)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['count'], 2)

    def test_node_list_multitenancy(self):
        path = reverse('node_list')
        org2 = self._create_org(name='org2')
        t2 = self._create_topology(organization=org2)
        self._create_node(
            label='node2', addresses=['192.168.0.2'], topology=t2, organization=org2
        )
        self.assertEqual(Node.objects.count(), 3)
        with self.assertNumQueries(6):
            response = self.client.get(path)
        self.assertEqual(response.status_code, 200)
        # Only nodes related to user's organization are returned
        self.assertEqual(response.data['count'], 2)

    def test_node_list_filter(self):
        admin = self._get_admin()
        self.client.force_login(admin)
        org1 = self._get_org()
        org2 = self._create_org(name='org2')
        t2 = self._create_topology(organization=org2)
        self._create_node(
            label='node2', addresses=['192.168.0.2'], topology=t2, organization=org2
        )
        path = reverse('node_list')

        with self.subTest('Test list without filters'):
            response = self.client.get(path)
            self.assertEqual(response.data['count'], 3)

        with self.subTest('Test filter by organization'):
            response = self.client.get(f'{path}?organization={org1.pk}')
            self.assertEqual(response.data['count'], 2)

        with self.subTest('Test filter by organization slug'):
            response = self.client.get(f'{path}?organization_slug={org1.slug}')
            self.assertEqual(response.data['count'], 2)

        with self.subTest('Test filter by topology'):
            response = self.client.get(f'{path}?topology={t2.pk}')
            self.assertEqual(response.data['count'], 1)

    def test_node_create_api(self):
        path = reverse('node_list')
        data = {
            'topology': self.topology.pk,
            'label': 'test-node',
            'addresses': ['192.168.0.1'],
            'properties': {},
            'user_properties': {},
        }
        with self.assertNumQueries(13):
            response = self.client.post(path, data, content_type='application/json')
        self.assertEqual(response.status_code, 201)
        self.assertEqual(response.data['topology'], self.topology.pk)
        self.assertEqual(response.data['label'], 'test-node')
        self.assertEqual(response.data['addresses'], ['192.168.0.1'])

    def test_node_create_api_shared_topology(self):
        admin = self._get_admin()
        self.client.force_login(admin)
        topology = self._create_topology(organization=None)
        org = self._get_org()
        data = {
            'topology': topology.pk,
            'label': 'test-node',
            'addresses': ['192.168.0.1'],
            'properties': {},
            'user_properties': {},
            'organization': org.pk,
        }
        path = reverse('node_list')
        with self.assertNumQueries(11):
            response = self.client.post(path, data, content_type='application/json')
        self.assertEqual(response.status_code, 201)
        self.assertEqual(response.data['topology'], topology.pk)
        self.assertEqual(response.data['organization'], org.pk)

    def test_node_detail_api(self):
        path = reverse('node_detail', args=(self.node1.pk,))
        with self.assertNumQueries(6):
            response = self.client.get(path)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['id'], str(self.node1.pk))
        self.assertEqual(response.data['label'], self.node1.label)
        self.assertEqual(response.data['topology'], self.node1.topology.pk)
        self.assertEqual(response.data['organization'], self.node1.organization.pk)
        self.assertEqual(response.data['addresses'], self.node1.addresses)
        self.assertEqual(response.data['properties'], self.node1.properties)
        self.assertEqual(response.data['user_properties'], self.node1.user_properties)
        self.assertIn('created', response.data)
        self.assertIn('modified', response.data)

    def test_node_put_api(self):
        data = {
            'topology': self.topology.pk,
            'label': 'change-node',
            'addresses': ['192.168.0.1', '192.168.0.2'],
            'properties': {},
            'user_properties': {},
        }
        path = reverse('node_detail', args=(self.node1.pk,))
        with self.assertNumQueries(13):
            response = self.client.put(path, data, content_type='application/json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['label'], 'change-node')
        self.assertEqual(response.data['addresses'], ['192.168.0.1', '192.168.0.2'])

    def test_node_patch_api(self):
        path = reverse('node_detail', args=(self.node1.pk,))
        data = {'label': 'change-node'}
        with self.assertNumQueries(12):
            response = self.client.patch(path, data, content_type='application/json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['label'], 'change-node')

    def test_node_delete_api(self):
        node = self._create_node(
            label='delete-node', addresses=['192.168.0.1'], topology=self.topology
        )
        path = reverse('node_detail', args=(node.pk,))
        response = self.client.delete(path)
        self.assertEqual(response.status_code, 204)
        self.assertEqual(Node.objects.filter(label='delete-node').count(), 0)

    def test_link_list_api(self):
        path = reverse('link_list')
        with self.assertNumQueries(6):
            response = self.client.get(path)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['count'], 1)

    def test_link_list_filters(self):
        admin = self._get_admin()
        self.client.force_login(admin)
        org1 = self._get_org()
        org2 = self._create_org(name='org2')
        t2 = self._create_topology(organization=org2)
        node1 = self._create_node(
            label='node2', addresses=['192.168.0.2'], topology=t2, organization=org2
        )
        node2 = self._create_node(
            label='node2', addresses=['192.168.0.3'], topology=t2, organization=org2
        )
        self._create_link(source=node1, target=node2, topology=t2, status='down')
        path = reverse('link_list')

        with self.subTest('Test list without filters'):
            response = self.client.get(path)
            self.assertEqual(response.data['count'], 2)

        with self.subTest('Test filter by organization'):
            response = self.client.get(f'{path}?organization={org1.pk}')
            self.assertEqual(response.data['count'], 1)

        with self.subTest('Test filter by organization slug'):
            response = self.client.get(f'{path}?organization_slug={org1.slug}')
            self.assertEqual(response.data['count'], 1)

        with self.subTest('Test filter by topology'):
            response = self.client.get(f'{path}?topology={t2.pk}')
            self.assertEqual(response.data['count'], 1)

        with self.subTest('Test filter by status'):
            response = self.client.get(f'{path}?status=down')
            self.assertEqual(response.data['count'], 1)

    def test_link_create_api(self):
        path = reverse('link_list')
        node3 = self._create_node(label='node3', topology=self.topology)
        data = {
            'topology': self.topology.pk,
            'source': self.node1.pk,
            'target': node3.pk,
            'cost': 1.0,
            'properties': {},
            'user_properties': {},
        }
        with self.assertNumQueries(16):
            response = self.client.post(path, data, content_type='application/json')
        self.assertEqual(response.status_code, 201)
        self.assertEqual(response.data['topology'], self.topology.pk)
        self.assertEqual(response.data['status'], 'up')
        self.assertEqual(response.data['source'], self.node1.pk)
        self.assertEqual(response.data['target'], node3.pk)

    def test_link_create_with_wrong_value_format_api(self):
        path = reverse('link_list')
        node3 = self._create_node(label='node3', topology=self.topology)
        data = {
            'topology': self.topology.pk,
            'source': self.node1.pk,
            'target': node3.pk,
            'cost': 1.0,
            'properties': 0,
            'user_properties': 122343,
        }
        with self.assertNumQueries(7):
            response = self.client.post(path, data, content_type='application/json')
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.data['properties'][0].title(),
            'Value Must Be Valid Json Or Key, Valued Pair.',
        )
        self.assertEqual(
            response.data['user_properties'][0].title(),
            'Value Must Be Valid Json Or Key, Valued Pair.',
        )

    def test_link_detail_api(self):
        path = reverse('link_detail', args=(self.link.pk,))
        with self.assertNumQueries(6):
            response = self.client.get(path)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['id'], str(self.link.pk))
        self.assertEqual(response.data['topology'], self.topology.pk)
        self.assertEqual(response.data['source'], self.node1.pk)
        self.assertEqual(response.data['target'], self.node2.pk)
        self.assertEqual(response.data['organization'], self.link.organization.pk)
        self.assertEqual(response.data['status'], self.link.status)
        self.assertEqual(response.data['cost'], self.link.cost)
        self.assertEqual(response.data['cost_text'], self.link.cost_text)
        self.assertEqual(response.data['properties'], self.link.properties)
        self.assertEqual(response.data['user_properties'], self.link.user_properties)
        self.assertIn('created', response.data)
        self.assertIn('modified', response.data)

    def test_link_put_api(self):
        path = reverse('link_detail', args=(self.link.pk,))
        data = {
            'topology': self.topology.pk,
            'source': self.node1.pk,
            'target': self.node2.pk,
            'cost': 21.0,
            'properties': {},
            'user_properties': {'user': 'tester'},
        }
        with self.assertNumQueries(16):
            response = self.client.put(path, data, content_type='application/json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['cost'], 21.0)
        self.assertEqual(response.data['properties'], {})
        self.assertEqual(response.data['user_properties'], {'user': 'tester'})

    def test_link_patch_api(self):
        path = reverse('link_detail', args=(self.link.pk,))
        data = {'cost': 50.0}
        with self.assertNumQueries(13):
            response = self.client.patch(path, data, content_type='application/json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['cost'], 50.0)

    def test_link_delete_api(self):
        self.assertEqual(Link.objects.count(), 1)
        path = reverse('link_detail', args=(self.link.pk,))
        response = self.client.delete(path)
        self.assertEqual(response.status_code, 204)
        self.assertEqual(Link.objects.count(), 0)



from openwisp_network_topology.api.views import LinkDetailView as BaseLinkDetailView
from openwisp_network_topology.api.views import (
    LinkListCreateView as BaseLinkListCreateView,
)
from openwisp_network_topology.api.views import (
    NetworkCollectionView as BaseNetworkCollectionView,
)
from openwisp_network_topology.api.views import (
    NetworkGraphHistoryView as BaseNetworkGraphHistoryView,
)
from openwisp_network_topology.api.views import NetworkGraphView as BaseNetworkGraphView
from openwisp_network_topology.api.views import NodeDetailView as BaseNodeDetailView
from openwisp_network_topology.api.views import (
    NodeListCreateView as BaseNodeListCreateView,
)
from openwisp_network_topology.api.views import (
    ReceiveTopologyView as BaseReceiveTopologyView,
)


class NetworkCollectionView(BaseNetworkCollectionView):
    pass


class NetworkGraphView(BaseNetworkGraphView):
    pass


class ReceiveTopologyView(BaseReceiveTopologyView):
    pass


class NetworkGraphHistoryView(BaseNetworkGraphHistoryView):
    pass


class NodeListCreateView(BaseNodeListCreateView):
    pass


class NodeDetailView(BaseNodeDetailView):
    pass


class LinkListCreateView(BaseLinkListCreateView):
    pass


class LinkDetailView(BaseLinkDetailView):
    pass


network_collection = NetworkCollectionView.as_view()
network_graph = NetworkGraphView.as_view()
network_graph_history = NetworkGraphHistoryView.as_view()
receive_topology = ReceiveTopologyView.as_view()
node_list = NodeListCreateView.as_view()
node_detail = NodeDetailView.as_view()
link_list = LinkListCreateView.as_view()
link_detail = LinkDetailView.as_view()



from collections import OrderedDict

import swapper
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from rest_framework import serializers

from openwisp_users.api.mixins import FilterSerializerByOrgManaged
from openwisp_utils.api.serializers import ValidatedModelSerializer

Node = swapper.load_model('topology', 'Node')
Link = swapper.load_model('topology', 'Link')
Topology = swapper.load_model('topology', 'Topology')


class NetworkCollectionSerializer(serializers.ListSerializer):
    """
    NetJSON NetworkCollection
    """

    @property
    def data(self):
        return super(serializers.ListSerializer, self).data

    def to_representation(self, data):
        return OrderedDict(
            (
                ('type', 'NetworkCollection'),
                ('collection', super().to_representation(data)),
            )
        )


def get_receive_url(pk, key):
    path = reverse('receive_topology', args=[pk])
    url = '{0}?key={1}'.format(path, key)
    return url


def get_representation_data(obj):
    """
    Returns a dict that represents
    a NetJSON NetworkGraph object.
    """
    nodes = []
    links = []
    for link in obj.get_links_queryset():
        links.append(link.json(dict=True, original=False))
    for node in obj.get_nodes_queryset():
        nodes.append(node.json(dict=True, original=False))
    netjson = OrderedDict(
        (
            ('type', 'NetworkGraph'),
            ('protocol', obj.protocol),
            ('version', obj.version),
            ('metric', obj.metric),
            ('label', obj.label),
            ('id', str(obj.id)),
            ('parser', obj.parser),
            ('organization', str(obj.organization.id) if obj.organization else None),
            ('strategy', obj.strategy),
            ('url', obj.url),
            ('key', obj.key),
            ('expiration_time', obj.expiration_time),
            ('receive_url', get_receive_url(obj.pk, obj.key)),
            ('published', obj.published),
            ('created', obj.created),
            ('modified', obj.modified),
            ('nodes', nodes),
            ('links', links),
        )
    )
    return netjson


class NetworkGraphRepresentation(object):
    def to_representation(self, obj):
        """
        Overriding the default represenation
        of Topology object.
        """
        topology_data = get_representation_data(obj)
        topology_data['receive_url'] = self.context['request'].build_absolute_uri(
            topology_data['receive_url']
        )
        return topology_data


class TopologySerializer(NetworkGraphRepresentation, ValidatedModelSerializer):
    class Meta:
        model = Topology
        fields = '__all__'


class BaseSerializer(FilterSerializerByOrgManaged, ValidatedModelSerializer):
    pass


class NetworkGraphSerializer(BaseSerializer):
    """
    NetJSON NetworkGraph.
    """

    def to_representation(self, obj):
        if self.context['request'].method == 'POST':
            serializer = TopologySerializer(
                self.instance, context={'request': self.context['request']}
            )
            return serializer.data
        return obj.json(dict=True)

    class Meta:
        model = Topology
        fields = (
            'label',
            'organization',
            'parser',
            'strategy',
            'key',
            'expiration_time',
            'url',
            'published',
        )
        list_serializer_class = NetworkCollectionSerializer
        extra_kwargs = {'published': {'initial': True}}


class NetworkGraphUpdateSerializer(NetworkGraphRepresentation, BaseSerializer):
    class Meta:
        model = Topology
        fields = (
            'label',
            'organization',
            'parser',
            'strategy',
            'key',
            'expiration_time',
            'url',
            'published',
        )

    def validate_strategy(self, value):
        if value == 'receive' and not self.initial_data.get('key'):
            raise serializers.ValidationError(
                _('A key must be specified when using RECEIVE strategy')
            )
        return value

    def validate_url(self, value):
        if not value and self.initial_data.get('strategy') == 'fetch':
            raise serializers.ValidationError(
                _('An url must be specified when using FETCH strategy')
            )
        return value


class BaseNodeLinkSerializer(BaseSerializer):
    properties = serializers.JSONField(initial={})

    def validate(self, data):
        instance = self.instance or self.Meta.model(**data)
        instance.full_clean()
        data['organization'] = instance.organization
        return data

    def validate_properties(self, value):
        if type(value) is not dict:
            raise serializers.ValidationError(
                _('Value must be valid JSON or key, valued pair.')
            )
        return value

    def validate_user_properties(self, value):
        if type(value) is not dict:
            raise serializers.ValidationError(
                _('Value must be valid JSON or key, valued pair.')
            )
        return value


class NodeSerializer(BaseNodeLinkSerializer):
    addresses = serializers.JSONField(initial=[])
    user_properties = serializers.JSONField(initial={})

    class Meta:
        model = Node
        fields = (
            'id',
            'topology',
            'organization',
            'label',
            'addresses',
            'properties',
            'user_properties',
            'created',
            'modified',
        )
        read_only_fields = ('created', 'modified')


class LinkSerializer(BaseNodeLinkSerializer):
    user_properties = serializers.JSONField(
        initial={},
        help_text=_('If you need to add additional data to this link use this field'),
    )

    class Meta:
        model = Link
        fields = (
            'id',
            'topology',
            'organization',
            'status',
            'source',
            'target',
            'cost',
            'cost_text',
            'properties',
            'user_properties',
            'created',
            'modified',
        )
        read_only_fields = ('organization', 'created', 'modified')



import logging
from ipaddress import ip_address, ip_network

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.utils.module_loading import import_string
from django.utils.timezone import datetime, now, timedelta
from django.utils.translation import gettext_lazy as _
from swapper import get_model_name, load_model

from openwisp_utils.base import UUIDModel

from .. import settings as app_settings

logger = logging.getLogger(__name__)

trigger_device_checks_path = 'openwisp_monitoring.device.tasks.trigger_device_checks'


class AbstractDeviceNode(UUIDModel):
    node = models.OneToOneField(
        get_model_name('topology', 'Node'), on_delete=models.CASCADE
    )
    device = models.ForeignKey(
        get_model_name('config', 'Device'), on_delete=models.CASCADE
    )

    # relations will be auto-created only for these parsers
    ENABLED_PARSERS = {
        'netdiff.OpenvpnParser': {
            'auto_create': 'auto_create_openvpn',
            'link_down': 'link_down_openvpn',
            'link_up': 'link_up_openvpn',
        },
        'netdiff.WireguardParser': {
            'auto_create': 'auto_create_wireguard',
        },
        'netdiff.ZeroTierParser': {
            'auto_create': 'auto_create_zerotier',
        },
        'netdiff.NetJsonParser': {
            'auto_create': 'auto_create_netjsongraph',
        },
    }

    class Meta:
        abstract = True
        unique_together = ('node', 'device')

    @classmethod
    def save_device_node(cls, device, node):
        device_node = cls(device=device, node=node)
        try:
            device_node.full_clean()
            device_node.save()
            # Update organization of the node. This is required
            # when topology is shared.
            if node.organization_id is None:
                node.organization_id = device.organization_id
                node.save(update_fields=['organization_id'])
        except Exception:
            logger.exception('Exception raised during save_device_node')
            return
        else:
            logger.info(f'DeviceNode relation created for {node.label} - {device.name}')
            return device_node

    @classmethod
    def auto_create(cls, node):
        """
        Attempts to perform automatic creation of DeviceNode objects.
        The right action to perform depends on the Topology used.
        """
        opts = cls.ENABLED_PARSERS.get(node.topology.parser)
        if opts:
            return getattr(cls, opts['auto_create'])(node)

    @classmethod
    def auto_create_openvpn(cls, node):
        """
        Implementation of the integration between
        controller and network-topology modules
        when using OpenVPN (using the common name)
        """
        common_name = node.properties.get('common_name')
        if not common_name:
            return

        Device = load_model('config', 'Device')
        device_filter = models.Q(config__vpnclient__cert__common_name=common_name)
        if node.organization_id:
            device_filter &= models.Q(organization_id=node.organization_id)
        device = (
            Device.objects.only(
                'id', 'name', 'last_ip', 'management_ip', 'organization_id'
            )
            .filter(device_filter)
            .first()
        )
        if not device:
            return

        return cls.save_device_node(device, node)

    @classmethod
    def auto_create_wireguard(cls, node):
        allowed_ips = node.properties.get('allowed_ips')
        if not allowed_ips:
            return
        Device = load_model('config', 'Device')
        ip_addresses = []
        for ip in allowed_ips:
            try:
                ip_addresses.extend([str(host) for host in ip_network(ip).hosts()])
            except ValueError:
                # invalid IP address
                continue
        device_filter = models.Q(config__vpnclient__ip__ip_address__in=ip_addresses)
        if node.organization_id:
            device_filter &= models.Q(organization_id=node.organization_id)
        device = (
            Device.objects.only(
                'id', 'name', 'last_ip', 'management_ip', 'organization_id'
            )
            .filter(device_filter)
            .first()
        )
        if not device:
            return
        return cls.save_device_node(device, node)

    @classmethod
    def auto_create_zerotier(cls, node):
        """
        Implementation of the integration between
        controller and network-topology modules
        when using ZeroTier (using the `zerotier_member_id`)
        """
        zerotier_member_id = node.properties.get('address')
        if not zerotier_member_id:
            return

        Device = load_model('config', 'Device')
        device_filter = models.Q(
            config__vpnclient__secret__startswith=zerotier_member_id
        )
        if node.organization_id:
            device_filter &= models.Q(organization_id=node.organization_id)
        device = (
            Device.objects.only(
                'id', 'name', 'last_ip', 'management_ip', 'organization_id'
            )
            .filter(device_filter)
            .first()
        )
        if not device:
            return
        return cls.save_device_node(device, node)

    @classmethod
    def auto_create_netjsongraph(cls, node):
        if len(node.addresses) < 2:
            # The first MAC address is Device.mac_address and
            # the second one is interface's MAC address.
            # If a node only has one MAC address, it means that
            # it only contains interface MAC address.
            return
        Device = load_model('config', 'Device')
        device_filter = models.Q(
            mac_address__iexact=node.addresses[0].rpartition('@')[0]
        )
        if node.organization_id:
            device_filter &= models.Q(organization_id=node.organization_id)
        device = (
            Device.objects.only(
                'id', 'name', 'last_ip', 'management_ip', 'organization_id'
            )
            .filter(device_filter)
            .first()
        )
        if not device:
            return
        return cls.save_device_node(device, node)

    def link_action(self, link, status):
        """
        Performs clean-up operations when link goes down.
        The right action to perform depends on the Topology used.
        """
        opts = self.ENABLED_PARSERS.get(link.topology.parser)
        if opts:
            key = f'link_{status}'
            if key in opts and hasattr(self, opts[key]):
                return getattr(self, opts[key])()

    def link_down_openvpn(self):
        """
        Link down action for OpenVPN
        """
        self.device.management_ip = None
        self.device.save()

    def link_up_openvpn(self):
        """
        Link up action for OpenVPN
        """
        addresses = self.node.addresses
        try:
            address = ip_address(addresses[1])
        except (IndexError, ValueError) as e:
            addresses = ', '.join(addresses)
            logger.warning(
                f'{e.__class__.__name__} raised while processing addresses: {addresses}'
            )
        else:
            self.device.management_ip = str(address)
            self.device.save()

    @classmethod
    def filter_by_link(cls, link):
        """
        Returns a queryset which looks for a DeviceNode which is related
        to the specified Link instance.
        """
        return cls.objects.filter(
            models.Q(node__source_link_set__pk=link.pk)
            | models.Q(node__target_link_set__pk=link.pk)
        ).select_related('device', 'node')

    @classmethod
    def trigger_device_updates(cls, link):
        """
        Used to refresh controller and monitoring information
        whenever the status of a link changes
        """
        if link.topology.parser not in cls.ENABLED_PARSERS:
            return
        for device_node in cls.filter_by_link(link):
            device_node.link_action(link, link.status)
            # triggers monitoring checks if OpenWISP Monitoring is enabled
            if 'openwisp_monitoring.device' in settings.INSTALLED_APPS:
                run_checks = import_string(trigger_device_checks_path)
                run_checks.delay(device_node.device.pk, recovery=link.status == 'up')


class AbstractWifiMesh(UUIDModel):
    _NODE_PROPERTIES = [
        'ht',
        'vht',
        'he',
        'mfp',
        'wmm',
        'vendor',
    ]
    _LINK_PROPERTIES = [
        'auth',
        'authorized',
        'noise',
        'signal',
        'signal_avg',
        'mesh_llid',
        'mesh_plid',
        'mesh_plink',
        'mesh_non_peer_ps',
    ]

    topology = models.ForeignKey(
        get_model_name('topology', 'Topology'), on_delete=models.CASCADE
    )
    mesh_id = models.CharField(
        max_length=32, null=False, blank=False, verbose_name=_('Mesh ID')
    )

    class Meta:
        abstract = True

    @classmethod
    def create_topology(cls, organization_ids, discard_older_data_time):
        if not app_settings.WIFI_MESH_INTEGRATION:
            raise ImproperlyConfigured(
                '"OPENIWSP_NETWORK_TOPOLOGY_WIFI_MESH_INTEGRATION" is set to "False".'
            )
        Link = load_model('topology', 'Link')
        for org_id in organization_ids:
            intermediate_topologies = cls._create_intermediate_topologies(
                org_id, discard_older_data_time
            )
            if not intermediate_topologies:
                Link.objects.filter(
                    topology__wifimesh__isnull=False, organization_id=org_id
                ).exclude(topology__wifimesh__in=intermediate_topologies.keys()).update(
                    status='down'
                )
                continue
            cls._create_topology(intermediate_topologies, org_id)

    @classmethod
    def _create_intermediate_topologies(cls, organization_id, discard_older_data_time):
        """
        Creates an intermediate data structure for creating topologies.
        The intermediate topology contains intermediate data structure
        for nodes and links.

        Every device in the mesh sends monitoring data. The data contains
        information of the clients the device is connected to.
        Using the information sent by individual device, a hub-spoke
        topology is created between device (hub) and clients (spoke).
        These individual topologies are then complied to create the
        complete mesh topology.
        """
        DeviceData = load_model('device_monitoring', 'DeviceData')
        intermediate_topologies = {}
        query = DeviceData.objects.filter(organization_id=organization_id).only(
            'mac_address'
        )
        discard_older_data_time = now() - timedelta(seconds=discard_older_data_time)
        for device_data in query.iterator():
            try:
                assert device_data.data is not None
                data_timestamp = datetime.fromisoformat(device_data.data_timestamp)
                assert data_timestamp > discard_older_data_time
            except (AttributeError, AssertionError):
                continue
            for interface in device_data.data.get('interfaces', []):
                if not AbstractWifiMesh._is_mesh_interfaces(interface):
                    continue
                mesh_id = '{}@{}'.format(
                    interface['wireless']['ssid'], interface['wireless']['channel']
                )
                if mesh_id not in intermediate_topologies:
                    intermediate_topologies[mesh_id] = {
                        'nodes': {},
                        'links': {},
                        'mac_mapping': {},
                    }
                topology = intermediate_topologies[mesh_id]
                (
                    collected_nodes,
                    collected_links,
                ) = cls._get_intermediate_nodes_and_links(
                    interface, device_data, topology['mac_mapping']
                )
                AbstractWifiMesh._merge_nodes(
                    interface, topology['nodes'], collected_nodes
                )
                AbstractWifiMesh._merge_links(
                    interface, topology['links'], collected_links
                )
        return intermediate_topologies

    @classmethod
    def _get_intermediate_nodes_and_links(cls, interface, device_data, mac_mapping):
        """
        Create intermediate data structures for nodes and links.
        These intermediate data structures are required because the
        interface's MAC address can be different from the device's main
        MAC address. Thus, these data structures are aimed to provide
        quick lookup while mapping interface MAC address to the
        device MAC address.
        """
        device_mac = device_data.mac_address.upper()
        interface_mac = interface['mac'].upper()
        channel = interface['wireless']['channel']
        device_node_id = f'{device_mac}@{channel}'
        mac_mapping[interface_mac] = device_node_id
        collected_nodes = {}
        collected_links = {}
        for client in interface['wireless'].get('clients', []):
            client_mac = client['mac'].upper()
            node_properties = {}
            for property in cls._NODE_PROPERTIES:
                if property in client:
                    node_properties[property] = client[property]
            collected_nodes[client_mac] = {'properties': node_properties}
            if client.get('mesh_plink') and client.get('mesh_plink') != 'ESTAB':
                # The link is not established.
                # Do not add this link to the topology.
                continue
            link_properties = {}
            for property in cls._LINK_PROPERTIES:
                if property in client:
                    link_properties[property] = client[property]

            collected_links[client_mac] = {
                interface_mac: {
                    'source': f'{device_mac}@{channel}',
                    'target': f'{client_mac}',
                    'cost': 1.0,
                    'properties': link_properties,
                }
            }
        return collected_nodes, collected_links

    @staticmethod
    def _is_mesh_interfaces(interface):
        return interface.get('wireless', False) and interface['wireless'].get(
            'mode'
        ) in ['802.11s']

    @staticmethod
    def _merge_nodes(interface, topology_nodes, collected_nodes):
        interface_mac = interface['mac'].upper()
        topology_nodes.update(collected_nodes)
        if not topology_nodes.get(interface_mac):
            # Handle case when there is only one node present
            # in the mesh
            topology_nodes[interface_mac] = {}

    @staticmethod
    def _merge_links(interface, topology_links, collected_links):
        """
        Merges properties of links from the topology stored
        in the database (topology_links) with link collected from
        the interface's wireless client data (collected_links).

        It takes into consideration the nature of the property
        (e.g. taking average for signal and noise, etc.).
        """
        interface_mac = interface['mac'].upper()
        if topology_links.get(interface_mac):
            for source, topology_link in topology_links[interface_mac].items():
                if not collected_links.get(source):
                    continue
                for property, value in collected_links[source][interface_mac][
                    'properties'
                ].items():
                    if isinstance(value, int):
                        # If the value is integer, then take average of the
                        # values provided by the two nodes.
                        # This is for fields like signal and noise.
                        if topology_link['properties'].get(property):
                            topology_link['properties'][property] = (
                                value + topology_link['properties'][property]
                            ) // 2
                        else:
                            # The link in topology does not contain this property,
                            # hence instead of averaging, assign the current value.
                            topology_link['properties'][property] = value
                    elif (
                        property in ['mesh_plink', 'mesh_non_peer_ps']
                        and topology_link['properties'].get(property)
                        and value != topology_link['properties'][property]
                    ):
                        # The value for "mesh_plink" and "mesh_non_peer_ps" properties
                        # should be reported same by both the nodes.
                        # Flag the value as inconsistent if they are different.
                        topology_link['properties'][
                            property
                        ] = 'INCONSISTENT: ({} / {})'.format(
                            value,
                            topology_link['properties'][property],
                        )
                collected_links.pop(source)
        for key, value in collected_links.items():
            try:
                topology_links[key].update(value)
            except KeyError:
                topology_links[key] = value

    @staticmethod
    def _create_topology(intermediate_topologies, organization_id):
        for mesh_id, intermediate in intermediate_topologies.items():
            topology_obj = AbstractWifiMesh._get_mesh_topology(mesh_id, organization_id)
            nodes = AbstractWifiMesh._get_nodes_from_intermediate_topology(intermediate)
            links = AbstractWifiMesh._get_links_from_intermediate_topology(intermediate)
            topology = {
                'type': 'NetworkGraph',
                'protocol': 'Mesh',
                'version': '1',
                'metric': 'Airtime',
                'nodes': nodes,
                'links': links,
            }
            topology_obj.receive(topology)

    @staticmethod
    def _get_nodes_from_intermediate_topology(intermediate_topology):
        """
        Using the "intermediate_topology", return all the
        nodes of the topology.

        It maps the interface's MAC address to the device's main MAC address
        which is used to create DeviceNode relation.
        """
        nodes = []
        mac_mapping = intermediate_topology['mac_mapping']
        for interface_mac, intermediate_node in intermediate_topology['nodes'].items():
            device_mac = mac_mapping.get(interface_mac)
            if not device_mac:
                continue
            node = {
                'id': device_mac,
                'label': device_mac,
                'local_addresses': [interface_mac],
            }
            if intermediate_node.get('properties'):
                node['properties'] = intermediate_node['properties']
            nodes.append(node)
        return nodes

    @staticmethod
    def _get_links_from_intermediate_topology(intermediate_topology):
        """
        Using the "intermediate_topology", return all the
        links of the topology.

        It maps the interface's MAC address to the device's main MAC address.
        """
        links = []
        mac_mapping = intermediate_topology['mac_mapping']
        for interface_mac, intermediate_link in intermediate_topology['links'].items():
            device_mac = mac_mapping.get(interface_mac)
            if not device_mac:
                continue
            for link in intermediate_link.values():
                link['target'] = device_mac
                links.append(link)
        return links

    @staticmethod
    def _get_mesh_topology(mesh_id, organization_id):
        """
        Get or create topology for the given mesh_id and organization_id.
        It also creates WifiMesh object to keep track of mesh's ID
        if a new topology object is created.
        """
        Topology = load_model('topology', 'Topology')
        WifiMesh = load_model('topology_device', 'WifiMesh')
        try:
            mesh_topology = (
                WifiMesh.objects.select_related('topology')
                .get(
                    mesh_id=mesh_id,
                    topology__organization_id=organization_id,
                )
                .topology
            )
        except WifiMesh.DoesNotExist:
            mesh_topology = Topology(
                organization_id=organization_id,
                label=mesh_id,
                parser='netdiff.NetJsonParser',
                strategy='receive',
                expiration_time=330,
            )
            mesh_topology.full_clean()
            mesh_topology.save()
            wifi_mesh = WifiMesh(
                mesh_id=mesh_id,
                topology=mesh_topology,
            )
            wifi_mesh.full_clean()
            wifi_mesh.save()
        return mesh_topology



import os

from django.conf import settings
from django.contrib import admin
from django.contrib.staticfiles.urls import staticfiles_urlpatterns
from django.urls import include, path

from openwisp_users.api.urls import get_api_urls

urlpatterns = []

if os.environ.get('SAMPLE_APP', False):
    from openwisp_network_topology.utils import get_visualizer_urls

    from .sample_network_topology.visualizer import views

    urlpatterns += [path('topology/', include(get_visualizer_urls(views)))]

urlpatterns += [
    path('', include('openwisp_network_topology.urls')),
    # needed to test integrations
    path('', include('openwisp_controller.urls')),
    path('admin/', admin.site.urls),
    path('api/v1/', include('openwisp_utils.api.urls')),
    path('api/v1/', include(get_api_urls())),
]

urlpatterns += staticfiles_urlpatterns()
if 'openwisp_monitoring.monitoring' in settings.INSTALLED_APPS:
    urlpatterns.append(path('', include('openwisp_monitoring.urls')))

if settings.DEBUG and 'debug_toolbar' in settings.INSTALLED_APPS:
    import debug_toolbar

    urlpatterns += [path('__debug__/', include(debug_toolbar.urls))]



from unittest.mock import patch
from uuid import uuid4

import pytest
from channels.db import database_sync_to_async
from channels.testing import WebsocketCommunicator
from django.conf import settings
from django.contrib.auth.models import Permission
from django.utils.module_loading import import_string
from swapper import load_model

from openwisp_users.tests.utils import TestOrganizationMixin

from .. import settings as app_settings
from .utils import CreateGraphObjectsMixin

Topology = load_model('topology', 'Topology')
Node = load_model('topology', 'Node')
Link = load_model('topology', 'Link')


@pytest.mark.asyncio
@pytest.mark.django_db(transaction=True)
class TestTopologySockets(CreateGraphObjectsMixin, TestOrganizationMixin):
    node_model = Node
    link_model = Link
    topology_model = Topology
    application = import_string(getattr(settings, 'ASGI_APPLICATION'))

    async def _get_communicator(self, admin_client, topology_id):
        session_id = admin_client.cookies['sessionid'].value
        communicator = WebsocketCommunicator(
            self.application,
            path=f'ws/network-topology/topology/{topology_id}/',
            headers=[
                (
                    b'cookie',
                    f'sessionid={session_id}'.encode('ascii'),
                )
            ],
        )
        return communicator

    async def _assert_connection_superuser(self, admin_client, conn=True):
        org = await database_sync_to_async(self._create_org)()
        t = await database_sync_to_async(self._create_topology)(organization=org)
        communicator = await self._get_communicator(admin_client, t.pk)
        connected, _ = await communicator.connect()
        assert connected is conn
        await communicator.disconnect()

    async def _assert_connection_org_manager(self, client, view_perm=True, conn=True):
        org = await database_sync_to_async(self._create_org)()
        test_user = await database_sync_to_async(self._create_user)(
            username='test-user-org-manager', email='test@orgmanger.com', is_staff=True
        )
        await database_sync_to_async(self._create_org_user)(
            is_admin=True, user=test_user, organization=org
        )
        t = await database_sync_to_async(self._create_topology)(organization=org)
        if view_perm:
            topology_view_permission = await database_sync_to_async(
                Permission.objects.get
            )(codename='view_topology')
            await database_sync_to_async(test_user.user_permissions.add)(
                topology_view_permission
            )
        await database_sync_to_async(client.force_login)(test_user)
        communicator = await self._get_communicator(client, t.pk)
        connected, _ = await communicator.connect()
        assert connected is conn
        await communicator.disconnect()

    async def _assert_connection_unauth_user(self, client, conn=True):
        client.cookies['sessionid'] = 'unauthenticated_user'
        org = await database_sync_to_async(self._create_org)()
        t = await database_sync_to_async(self._create_topology)(organization=org)
        communicator = await self._get_communicator(client, t.pk)
        connected, _ = await communicator.connect()
        assert connected is conn
        await communicator.disconnect()

    async def test_consumer_connection_superuser(self, admin_client):
        await self._assert_connection_superuser(admin_client)

    async def test_consumer_connection_org_manager_view_perm(self, client):
        await self._assert_connection_org_manager(client)

    async def test_consumer_connection_org_manager_without_view_perm(self, client):
        await self._assert_connection_org_manager(client, view_perm=False, conn=False)

    async def test_consumer_connection_unauth_user(self, client):
        await self._assert_connection_unauth_user(client, conn=False)

    @patch.object(app_settings, 'TOPOLOGY_API_AUTH_REQUIRED', False)
    async def test_consumer_connection_auth_disabled_superuser(self, admin_client):
        await self._assert_connection_superuser(admin_client)

    @patch.object(app_settings, 'TOPOLOGY_API_AUTH_REQUIRED', False)
    async def test_consumer_connection_auth_disabled_org_manager(self, client):
        await self._assert_connection_org_manager(client)

    @patch.object(app_settings, 'TOPOLOGY_API_AUTH_REQUIRED', False)
    async def test_consumer_connection_auth_disabled_unauth_user(self, client):
        await self._assert_connection_unauth_user(client)

    async def test_consumer_connection_unexisting_topology(
        self, admin_user, admin_client
    ):
        org = await database_sync_to_async(self._create_org)()
        await database_sync_to_async(self._create_topology)(organization=org)
        communicator = await self._get_communicator(admin_client, uuid4())
        connected, _ = await communicator.connect()
        assert connected is False
        await communicator.disconnect()

    async def test_node_topology_update(self, admin_user, admin_client):
        org = await database_sync_to_async(self._create_org)()
        topo = await database_sync_to_async(self._create_topology)(organization=org)
        communicator = await self._get_communicator(admin_client, topo.pk)
        connected, _ = await communicator.connect()
        assert connected is True
        node = await database_sync_to_async(self._create_node)(
            topology=topo, organization=org
        )
        response = await communicator.receive_json_from()
        expected_response = await database_sync_to_async(topo.json)()
        assert response['topology'] is not None
        assert response['topology'] == expected_response
        node.label = 'test'
        await database_sync_to_async(node.full_clean)()
        await database_sync_to_async(node.save)()
        expected_response = await database_sync_to_async(topo.json)()
        response = await communicator.receive_json_from()
        assert response['topology'] is not None
        assert response['topology'] == expected_response
        await database_sync_to_async(node.delete)()
        expected_response = await database_sync_to_async(topo.json)()
        response = await communicator.receive_json_from()
        assert response['topology'] is not None
        assert response['topology'] == expected_response
        await communicator.disconnect()

    async def test_link_topology_update(self, admin_user, admin_client):
        org = await database_sync_to_async(self._create_org)()
        topo = await database_sync_to_async(self._create_topology)(organization=org)
        communicator = await self._get_communicator(admin_client, topo.pk)
        connected, _ = await communicator.connect()
        assert connected is True
        node1 = await database_sync_to_async(self._create_node)(
            topology=topo, label='node-0', organization=org
        )
        response = await communicator.receive_json_from()
        node2 = await database_sync_to_async(self._create_node)(
            topology=topo, label='node-1', organization=org
        )
        response = await communicator.receive_json_from()
        link = await database_sync_to_async(self._create_link)(
            topology=topo, source=node1, target=node2, organization=org
        )
        response = await communicator.receive_json_from()
        expected_response = await database_sync_to_async(topo.json)()
        assert response['topology'] is not None
        assert response['topology'] == expected_response
        link.status = 'down'
        await database_sync_to_async(link.full_clean)()
        await database_sync_to_async(link.save)()
        expected_response = await database_sync_to_async(topo.json)()
        response = await communicator.receive_json_from()
        assert response['topology'] is not None
        assert response['topology'] == expected_response
        await database_sync_to_async(link.delete)()
        expected_response = await database_sync_to_async(topo.json)()
        response = await communicator.receive_json_from()
        assert response['topology'] is not None
        assert response['topology'] == expected_response
        await communicator.disconnect()

    async def test_topology_properties_update(self, admin_user, admin_client):
        org = await database_sync_to_async(self._create_org)()
        topo = await database_sync_to_async(self._create_topology)(organization=org)
        communicator = await self._get_communicator(admin_client, topo.pk)
        connected, _ = await communicator.connect()
        assert connected is True
        topo.name = 'test'
        await database_sync_to_async(topo.full_clean)()
        await database_sync_to_async(topo.save)()
        expected_response = await database_sync_to_async(topo.json)()
        response = await communicator.receive_json_from()
        assert response['topology'] is not None
        assert response['topology'] == expected_response
        await communicator.disconnect()



import swapper
from openwisp_ipam.tests import CreateModelsMixin as SubnetIpamMixin

from openwisp_controller.config.tests.utils import (
    CreateConfigTemplateMixin,
    TestVpnX509Mixin,
    TestWireguardVpnMixin,
)
from openwisp_network_topology.tests.utils import CreateGraphObjectsMixin
from openwisp_users.tests.utils import TestOrganizationMixin

Node = swapper.load_model('topology', 'Node')

Topology = swapper.load_model('topology', 'Topology')


class TopologyTestMixin(
    TestVpnX509Mixin,
    TestWireguardVpnMixin,
    SubnetIpamMixin,
    CreateConfigTemplateMixin,
    CreateGraphObjectsMixin,
    TestOrganizationMixin,
):
    topology_model = Topology
    node_model = Node

    def _init_test_node(
        self,
        topology,
        addresses=None,
        label='test',
        common_name=None,
        create=True,
    ):
        if not addresses:
            addresses = ['netjson_id']
        node = Node(
            organization=topology.organization,
            topology=topology,
            label=label,
            addresses=addresses,
            properties={'common_name': common_name},
        )
        if create:
            node.full_clean()
            node.save()
        return node

    def _init_wireguard_test_node(self, topology, addresses=[], create=True, **kwargs):
        if not addresses:
            addresses = ['public_key']
        properties = {
            'preshared_key': None,
            'endpoint': None,
            'latest_handsake': '0',
            'transfer_rx': '0',
            'transfer_tx': '0',
            'persistent_keepalive': 'off',
            'allowed_ips': ['10.0.0.2/32'],
        }
        properties.update(kwargs)
        allowed_ips = properties.get('allowed_ips')
        node = Node(
            organization=topology.organization,
            topology=topology,
            label=','.join(allowed_ips),
            addresses=addresses,
            properties=properties,
        )
        if create:
            node.full_clean()
            node.save()
        return node

    def _create_wireguard_test_env(self, parser):
        org = self._get_org()
        device, _, _ = self._create_wireguard_vpn_template()
        device.organization = org
        topology = self._create_topology(organization=org, parser=parser)
        return topology, device

    def _create_test_env(self, parser):
        organization = self._get_org()
        vpn = self._create_vpn(name='test VPN', organization=organization)
        self._create_template(
            name='VPN',
            type='vpn',
            vpn=vpn,
            config=vpn.auto_client(),
            default=True,
            organization=organization,
        )
        vpn2 = self._create_vpn(name='test VPN2', ca=vpn.ca, organization=organization)
        self._create_template(
            name='VPN2',
            type='vpn',
            vpn=vpn2,
            config=vpn.auto_client(),
            default=True,
            organization=organization,
        )
        device = self._create_device(organization=organization)
        config = self._create_config(device=device)
        topology = self._create_topology(organization=organization, parser=parser)
        cert = config.vpnclient_set.first().cert
        return topology, device, cert



from django.contrib.staticfiles.testing import StaticLiveServerTestCase
from django.test import tag
from django.urls import reverse
from selenium.common.exceptions import (
    ElementClickInterceptedException,
    TimeoutException,
)
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait
from swapper import load_model

from openwisp_users.tests.utils import TestOrganizationMixin
from openwisp_utils.test_selenium_mixins import SeleniumTestMixin

from .utils import CreateGraphObjectsMixin, LoadMixin

Link = load_model('topology', 'Link')
Node = load_model('topology', 'Node')
Topology = load_model('topology', 'Topology')


@tag('selenium_tests')
class TestTopologyGraphVisualizer(
    TestOrganizationMixin,
    CreateGraphObjectsMixin,
    LoadMixin,
    SeleniumTestMixin,
    StaticLiveServerTestCase,
):
    app_label = 'topology'
    node_model = Node
    link_model = Link
    topology_model = Topology
    _EXPECTED_KEYS = ['Label', 'Protocol', 'Version', 'Metric', 'Nodes', 'Links']

    @property
    def prefix(self):
        return f'admin:{self.app_label}'

    def setUp(self):
        org = self._create_org()
        self.admin = self._create_admin(
            username=self.admin_username, password=self.admin_password
        )
        self.topology = self._create_topology(organization=org)
        self._create_node(
            label='node1',
            addresses=['192.168.0.1'],
            topology=self.topology,
            organization=org,
        )
        self._create_node(
            label='node2',
            addresses=['192.168.0.2'],
            topology=self.topology,
            organization=org,
        )

    def tearDown(self):
        # Clear the web driver console logs after each test
        self.web_driver.get_log('browser')

    def _get_console_errors(self, console_logs=[]):
        console_errors = []
        exclude_socket_icon_errors = [
            'WebSocket handshake',
            'favicon.ico - Failed to load resource:',
        ]
        for log in console_logs:
            error_message = log['message']
            error_level = log['level']
            if error_level == 'SEVERE' and any(
                [err in error_message for err in exclude_socket_icon_errors]
            ):
                continue
            console_errors.append(log)
        return console_errors

    def _assert_topology_graph(self):
        WebDriverWait(self.web_driver, 2).until(
            EC.invisibility_of_element_located((By.ID, 'loadingContainer'))
        )
        sidebar = WebDriverWait(self.web_driver, 2).until(
            EC.visibility_of_element_located((By.CLASS_NAME, 'sideBarHandle'))
        )
        sidebar.click()
        WebDriverWait(self.web_driver, 2).until(
            EC.visibility_of_element_located((By.CLASS_NAME, 'njg-metaInfoContainer'))
        )
        try:
            # Check the visibility of the topology graph
            WebDriverWait(self.web_driver, 2).until(
                EC.visibility_of_element_located((By.XPATH, '//div[1]/canvas'))
            )
        except TimeoutException:
            self.fail('The topology graph did not render')

        console_logs = self.web_driver.get_log('browser')
        console_errors = self._get_console_errors(console_logs)
        self.assertEqual(console_errors, [])
        topology_graph_dict = self.topology.json(dict=True)
        topology_graph_label_keys = self.web_driver.find_elements(
            By.CSS_SELECTOR, '.njg-keyLabel'
        )
        topology_graph_label_values = self.web_driver.find_elements(
            By.CSS_SELECTOR, '.njg-valueLabel'
        )
        self.assertEqual(len(topology_graph_label_keys), len(self._EXPECTED_KEYS))
        # ensure correct topology graph labels are present
        for key in topology_graph_label_keys:
            self.assertIn(key.text, self._EXPECTED_KEYS)
        expected_label_values = [
            topology_graph_dict['label'],
            topology_graph_dict['protocol'],
            topology_graph_dict['version'],
            topology_graph_dict['metric'],
            str(len(topology_graph_dict['nodes'])),
            str(len(topology_graph_dict['links'])),
        ]
        for label_value, expected_value in zip(
            topology_graph_label_values, expected_label_values
        ):
            self.assertEqual(label_value.text, expected_value)

    def test_topology_admin_view_graph_visualizer(self):
        path = reverse(f'{self.prefix}_topology_change', args=[self.topology.pk])
        self.login(username=self.admin_username, password=self.admin_password)
        self.open(path)
        self.web_driver.find_element(By.CSS_SELECTOR, 'input.visualizelink').click()
        self._assert_topology_graph()

    def test_unpublished_topology_admin_view_graph_visualizer(self):
        self.topology_model.objects.update(published=False)
        path = reverse(f'{self.prefix}_topology_change', args=[self.topology.pk])
        self.login(username=self.admin_username, password=self.admin_password)
        self.open(path)
        self.web_driver.find_element(By.CSS_SELECTOR, 'input.visualizelink').click()
        self._assert_topology_graph()

    def test_topology_non_admin_view_graph_visualizer(self):
        path = reverse('topology_list')
        self.login(username=self.admin_username, password=self.admin_password)
        self.web_driver.get(f'{self.live_server_url}{path}')
        topology_graph_element = self.web_driver.find_element(
            By.XPATH, "//ul[@id='menu']/li/a"
        )
        topology_graph_element.click()
        self._assert_topology_graph()

    def test_topology_admin_visualizer_multiple_close_btn_append(self):
        path = reverse(f'{self.prefix}_topology_change', args=[self.topology.pk])
        self.login(username=self.admin_username, password=self.admin_password)
        self.open(path)
        self.web_driver.find_element(By.CSS_SELECTOR, 'input.visualizelink').click()
        self._assert_topology_graph()
        self.web_driver.find_element(By.CLASS_NAME, 'closeBtn').click()
        # Now try to open visualizer
        # again and make sure only single
        # 'closeBtn' element is present in the DOM
        self.web_driver.find_element(By.CSS_SELECTOR, 'input.visualizelink').click()
        self._assert_topology_graph()
        try:
            self.web_driver.find_element(By.CLASS_NAME, 'closeBtn').click()
        except ElementClickInterceptedException:
            self.fail('Multiple "closeBtn" are present in the visualizer DOM')

    def test_topology_admin_esc_key_close_visualizer(self):
        path = reverse(f'{self.prefix}_topology_change', args=[self.topology.pk])
        self.login(username=self.admin_username, password=self.admin_password)
        self.open(path)
        self.web_driver.find_element(By.CSS_SELECTOR, 'input.visualizelink').click()
        self._assert_topology_graph()
        # Try to close the visualizer with the "Esc" key.
        body = self.web_driver.find_element(By.TAG_NAME, 'body')
        body.send_keys(Keys.ESCAPE)
        # Open the visualizer again and make sure no JS errors
        # are thrown when the visualizer is closed again
        self.web_driver.find_element(By.CSS_SELECTOR, 'input.visualizelink').click()
        self._assert_topology_graph()
        self.web_driver.find_element(By.CLASS_NAME, 'closeBtn').click()
        console_logs = self.web_driver.get_log('browser')
        console_errors = self._get_console_errors(console_logs)
        self.assertEqual(console_errors, [])



# Generated by Django 3.2.18 on 2023-04-17 18:06

import uuid

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.TOPOLOGY_TOPOLOGY_MODEL),
        ('topology_device', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='WifiMesh',
            fields=[
                (
                    'id',
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ('mesh_id', models.CharField(max_length=32, verbose_name='Mesh ID')),
                (
                    'topology',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.TOPOLOGY_TOPOLOGY_MODEL,
                    ),
                ),
            ],
            options={
                'abstract': False,
                'swappable': 'TOPOLOGY_DEVICE_WIFIMESH_MODEL',
            },
        ),
    ]



import json
from copy import deepcopy
from unittest.mock import patch

import swapper
from django.core.cache import cache
from django.core.exceptions import ImproperlyConfigured
from django.test import TransactionTestCase, tag
from django.urls import reverse
from django.utils.timezone import now

from .. import settings as app_settings
from ..tasks import create_mesh_topology
from . import SIMPLE_MESH_DATA, SINGLE_NODE_MESH_DATA
from .utils import TopologyTestMixin

Node = swapper.load_model('topology', 'Node')
Link = swapper.load_model('topology', 'Link')
Topology = swapper.load_model('topology', 'Topology')
DeviceNode = swapper.load_model('topology_device', 'DeviceNode')
WifiMesh = swapper.load_model('topology_device', 'WifiMesh')
Device = swapper.load_model('config', 'Device')


@tag('wifi_mesh')
class TestWifiMeshIntegration(TopologyTestMixin, TransactionTestCase):
    app_label = 'topology'

    def setUp(self):
        super().setUp()
        cache.clear()

    @property
    def prefix(self):
        return 'admin:{0}'.format(self.app_label)

    def _populate_mesh(self, data):
        org = self._get_org()
        devices = []
        for mac, interfaces in data.items():
            device = self._create_device(name=mac, mac_address=mac, organization=org)
            devices.append(device)
            response = self.client.post(
                '{0}?key={1}&time={2}'.format(
                    reverse('monitoring:api_device_metric', args=[device.id]),
                    device.key,
                    now().utcnow().strftime('%d-%m-%Y_%H:%M:%S.%f'),
                ),
                data=json.dumps(
                    {
                        'type': 'DeviceMonitoring',
                        'interfaces': interfaces,
                    }
                ),
                content_type='application/json',
            )
            self.assertEqual(response.status_code, 200)
        create_mesh_topology.delay(organization_ids=(org.id,))
        return devices, org

    @patch.object(app_settings, 'WIFI_MESH_INTEGRATION', False)
    def test_wifi_mesh_integration_disabled(self):
        with self.subTest('Test calling "create_mesh_topology" task'):
            with patch.object(WifiMesh, 'create_topology') as mocked:
                _, org = self._populate_mesh(SIMPLE_MESH_DATA)
                self.assertEqual(Topology.objects.count(), 0)
                mocked.assert_not_called()

        # Ensure the following sub-test does not fail if the
        # previous one fails.
        Topology.objects.all().delete()

        with self.subTest('Test calling WifiMesh.create_topology'):
            with self.assertRaises(ImproperlyConfigured) as error:
                WifiMesh.create_topology(
                    organization_ids=[org.id], discard_older_data_time=360
                )
            self.assertEqual(
                str(error.exception),
                '"OPENIWSP_NETWORK_TOPOLOGY_WIFI_MESH_INTEGRATION" is set to "False".',
            )

    def test_simple_mesh(self):
        devices, org = self._populate_mesh(SIMPLE_MESH_DATA)
        self.assertEqual(Topology.objects.filter(organization=org).count(), 1)
        topology = Topology.objects.filter(organization=org).first()
        self.assertEqual(
            WifiMesh.objects.filter(topology=topology, mesh_id='Test Mesh@11').count(),
            1,
        )
        self.assertEqual(
            Node.objects.filter(
                topology=topology,
                organization=org,
                properties__contains=(
                    '{\n    "ht": true,\n    "vht": null,\n    "mfp": false,\n'
                    '    "wmm": true,\n    "vendor": "TP-LINK TECHNOLOGIES CO.,LTD."\n}'
                ),
            ).count(),
            3,
        )
        self.assertEqual(
            Link.objects.filter(
                topology=topology,
                organization=org,
                properties__contains='"noise": -94',
            )
            .filter(properties__contains='"signal": -58')
            .filter(properties__contains='"mesh_llid": 19500')
            .filter(properties__contains='"mesh_plid": 24500')
            .count(),
            3,
        )
        self.assertEqual(
            Link.objects.filter(
                topology=topology,
                organization=org,
                properties__contains='"mesh_non_peer_ps": "INCONSISTENT: (LISTEN / ACTIVE)"',
            ).count(),
            1,
        )
        self.assertEqual(DeviceNode.objects.filter(device__in=devices).count(), 3)

        # Test DeviceNode creation logic is not executed when the create_topology
        # is executed again
        with patch.object(DeviceNode, 'auto_create') as mocked_auto_create:
            create_mesh_topology.delay(organization_ids=(org.id,))
            mocked_auto_create.assert_not_called()

    @patch('logging.Logger.exception')
    def test_single_node_mesh(self, mocked_logger):
        devices, org = self._populate_mesh(SINGLE_NODE_MESH_DATA)
        self.assertEqual(Topology.objects.filter(organization=org).count(), 1)
        topology = Topology.objects.filter(organization=org).first()
        self.assertEqual(
            WifiMesh.objects.filter(topology=topology, mesh_id='Test Mesh@11').count(),
            1,
        )
        self.assertEqual(
            Node.objects.filter(
                topology=topology,
                organization=org,
            ).count(),
            1,
        )
        self.assertEqual(
            Link.objects.filter(
                topology=topology,
                organization=org,
            ).count(),
            0,
        )
        self.assertEqual(DeviceNode.objects.filter(device__in=devices).count(), 1)
        mocked_logger.assert_not_called()

    def test_mesh_id_changed(self):
        devices, org = self._populate_mesh(SIMPLE_MESH_DATA)
        self.assertEqual(Topology.objects.filter(organization=org).count(), 1)
        self.assertEqual(WifiMesh.objects.filter(mesh_id='Test Mesh@11').count(), 1)
        topology = Topology.objects.filter(organization=org).first()
        self.assertEqual(
            Node.objects.filter(
                topology=topology,
                organization=org,
            ).count(),
            3,
        )
        self.assertEqual(
            Link.objects.filter(
                topology=topology,
                organization=org,
            ).count(),
            3,
        )
        # Change mesh_id reported in the monitoring data
        mesh_data = deepcopy(SIMPLE_MESH_DATA)
        for device, interfaces in zip(devices, mesh_data.values()):
            interfaces[0]['wireless']['ssid'] = 'New Mesh'
            response = self.client.post(
                '{0}?key={1}&time={2}'.format(
                    reverse('monitoring:api_device_metric', args=[device.id]),
                    device.key,
                    now().utcnow().strftime('%d-%m-%Y_%H:%M:%S.%f'),
                ),
                data=json.dumps(
                    {
                        'type': 'DeviceMonitoring',
                        'interfaces': interfaces,
                    }
                ),
                content_type='application/json',
            )
            self.assertEqual(response.status_code, 200)
        create_mesh_topology.delay(organization_ids=(org.id,))
        self.assertEqual(Topology.objects.filter(organization=org).count(), 2)
        self.assertEqual(WifiMesh.objects.count(), 2)
        self.assertEqual(Node.objects.count(), 6)
        self.assertEqual(Link.objects.count(), 6)
        self.assertEqual(WifiMesh.objects.filter(mesh_id='New Mesh@11').count(), 1)
        topology = Topology.objects.filter(
            organization=org, wifimesh__mesh_id='New Mesh@11'
        ).first()
        self.assertEqual(
            Node.objects.filter(
                topology=topology,
                organization=org,
            ).count(),
            3,
        )
        self.assertEqual(
            Link.objects.filter(
                topology=topology,
                organization=org,
            ).count(),
            3,
        )

    def test_topology_admin(self):
        """
        Tests WifiMeshInlineAdmin is present in TopologyAdmin
        when OPENWISP_NETWORK_TOPOLOGY_WIFI_MESH_INTEGRATION
        is set to True.

        Note: This test is present here because TopologyAdmin class
        cannot be patched based on app_settings.WIFI_MESH_INTEGRATION
        once the project is initialized.
        """
        admin = self._create_admin()
        self.client.force_login(admin)

        with self.subTest('Test add form'):
            response = self.client.get(reverse(f'{self.prefix}_topology_add'))
            self.assertContains(response, 'Wifi mesh')

        with self.subTest('Test change form'):
            topology = self._create_topology()
            response = self.client.get(
                reverse(f'{self.prefix}_topology_change', args=[topology.id])
            )
            self.assertContains(response, 'Wifi mesh')



from django.conf import settings
from django.contrib.auth.models import Permission
from django.db import migrations

from openwisp_network_topology.migrations import create_default_permissions


def assign_permissions_to_groups(apps, schema_editor):
    create_default_permissions(apps, schema_editor)

    def _add_permission_to_group(group, models, operations):
        for model in models:
            for operation in operations:
                try:
                    permission = Permission.objects.get(
                        codename='{}_{}'.format(operation, model)
                    )
                except Permission.DoesNotExist:
                    continue
                else:
                    group.permissions.add(permission.pk)

    Group = apps.get_model('openwisp_users', 'Group')
    manage_operations = ['add', 'change', 'delete', 'view']
    try:
        admin = Group.objects.get(name='Administrator')
        operator = Group.objects.get(name='Operator')
    # consider failures custom cases
    # that do not have to be dealt with
    except Group.DoesNotExist:
        return

    _add_permission_to_group(operator, ['wifimesh'], manage_operations)
    _add_permission_to_group(admin, ['wifimesh'], manage_operations)


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.TOPOLOGY_TOPOLOGY_MODEL),
        ('topology_device', '0002_wifimesh'),
    ]

    operations = [
        migrations.RunPython(
            assign_permissions_to_groups, reverse_code=migrations.RunPython.noop
        )
    ]



from swapper import load_model

from openwisp_users.api.filters import OrganizationManagedFilter

Node = load_model('topology', 'Node')
Link = load_model('topology', 'Link')
Topology = load_model('topology', 'Topology')


class NetworkCollectionFilter(OrganizationManagedFilter):
    class Meta(OrganizationManagedFilter.Meta):
        model = Topology
        fields = OrganizationManagedFilter.Meta.fields + ['strategy', 'parser']


class NodeFilter(OrganizationManagedFilter):
    class Meta(OrganizationManagedFilter.Meta):
        model = Node
        fields = OrganizationManagedFilter.Meta.fields + ['topology']


class LinkFilter(OrganizationManagedFilter):
    class Meta(OrganizationManagedFilter.Meta):
        model = Link
        fields = OrganizationManagedFilter.Meta.fields + ['topology', 'status']
